/*
	Version	0.1	// 25.09.2020
	
	USAGE:
	
		set	procedure to ExportData.cc additive
		o	= new ExportData()
		
		o.SetSource(type, rowsetObject|tableName|fileName|sheetName|SELECT ..., databaseObject|dataBaseName|folder|xlsxFileName)
		
		o.SetDestination(type, rowsetObject|tableName|fileName|sheetName|SELECT ..., databaseObject|dataBaseName|folder|xlsxFileName)
		
		type	can be:	"TXT", "CSV", "TAB", "TEXT", "DELIMITED"		text file
							"XLSX", "XLS"											excel file
							"SQL", "DBF", "BDE",  null							dbf, sql
							"ADO"														adoDatabase
							
		o.Copy()				//CREATE Destination Table with structure from Source table
		
		o.append()			//APPEND rows in Destination Table from Source table
	
	Examples:
		Export DBF to XLSX
				o	= new ExportData()
				o.setSource('DBF', '[path\]myTable')
				o.setDestination('XLSX', 'mySheet', '[path\]myXlsFile.xslx')
				o.copy()
				o.release()

		Export ADO to XLSX
				o	= new ExportData()
				o.setSource('ADO', 'myTable', myConnectionString)						myConnectionString must include Server, UserName, Password if requided
				o.setDestination('XLSX', 'mySheet', '[path\]myXlsFile.xslx')
				o.copy()
				o.release()
			
		Export CSV to XLSX
				o	= new ExportData()
				o.setSource('CSV', 'myFile', pathToMyFile)
				o.setDestination('XLSX', 'mySheet', '[path\]myXlsFile.xslx')
				
		FOR CSV, TXT files
			can be set additional option if required, by ex.
				o.AddSchemaOptions(o.Source, 'DecimalSymbol',	',')
				o.AddSchemaOptions(o.Source, 'Format',				'Delimited(;)'	)
				etc ...
				
			MUST set fields description, by ex.
				o.AddSchemaColumns(o.Source, 'Col1', 'MyFieldA',	'LONG')
				o.AddSchemaColumns(o.Source, 'Col2', 'MyFieldB',	'CHAR WIDTH 30')
				...
				o.AddSchemaColumns(o.Source, 'Col10', 'MyFieldX',	'DATE')
				...

//
//			For additional information about SCHEMA.INI please read:
//
//		https://docs.microsoft.com/en-us/sql/odbc/microsoft/schema-ini-file-text-file-driver?view=sql-server-ver15
//
				
				o.copy()
				o.release()
				

//
//			XLSX files require MS AccessRunTime to be instaled, can be downloaded from:
//			
//		https://www.microsoft.com/en-us/download/details.aspx?id=50040
//

*/



#define	isCSV(x)			({"TXT", "CSV", "TAB", "TEXT", "DELIMITED"}.scan(x) > 0)
#define	isXLS(x)			({"XLSX", "XLS"}.scan(x) > 0)
#define	isADO(x)			({"ADO"}.scan(x) > 0)
#define	isLOCALSQL(x)	({"SQL", "DBF", "BDE",  null}.scan(x) > 0)
#define	isAdoType(x)	(isADO(X) OR isCSV(x) OR isXLS(x))
#define	isSqlSelect(x)	{|xPar| ; private c ; c=xPar ; return type('c')=='C' and upper(c) = "SELECT " }(x)


CLASS ExportData
PROTECT	CreateSchemaIni, CopyData, Inactivate

set	procedure to :duflp:ini.cc additive

this.Source				= new object()
this.Source.Name		= 'SOURCE'
this.Source.Type		= null
this.Source.Table		= null
this.Source.DataBase	= null

//
//			For information about SCHEMA.INI please read:
//
//		https://docs.microsoft.com/en-us/sql/odbc/microsoft/schema-ini-file-text-file-driver?view=sql-server-ver15
//

	this.Source.Schema = new object()		
	this.Source.Schema.DefaultOptions	= new array()
	this.Source.Schema.Options 			= new array()
	this.Source.Schema.Columns 			= new array()



	
this.Destination		= new object()
this.Destination.Type		= null
this.Destination.Name		= 'DESTINATION'
this.Destination.Table		= null
this.Destination.DataBase	= null

	this.Destination.Schema	= new object()		
	this.Destination.Schema.DefaultOptions = new array()
	this.Destination.Schema.Options 			= new array()
	this.Destination.Schema.Columns			= new array()
	
	
with	this.Source.Schema
	DefaultOptions.add({'Format','CSVDelimited'})
	DefaultOptions.add({'ColNameHeader', false })
	DefaultOptions.add({'MaxScanRows',1})
endwith
	
with	this.Destination.Schema
	DefaultOptions.add({'Format','CSVDelimited'})
	DefaultOptions.add({'ColNameHeader', false })
	DefaultOptions.add({'TextDelimiter', 'none'})
endwith



/*************		Methods	*************/


//  Parameters for SetSource, SetDestination				//pType	= ADO;  XLSX,XLS;  TXT,CSV,TEXT;  DBF,SQL,BDE
																		//pTable	= rowset object  ; sheetName; fileName; tableNamesql; "SELECT ..."
																		//pDatabase	=  activated ADOdatabase Object or database Object		OR
																		//					for ADO				connectionstring with defined Server, User, Password
																		//					for TXT,CSV,TEXT	folder(directory) which contain the file
																		//					for xlsx,xls		(path +) excel filename
																		//					for dbf,sql,bde	databaseName OR null (default database)
FUNCTION SetSource( pType, pTable, pDatabase )

	this.Source.Type	:= UPPER(pType)
	this.Source.Table	:= pTable

	if	argCount() >= 3
		this.Source.Database	:= pDatabase
	else
		this.Source.Database	:= null
	endif

	return



FUNCTION SetDestination( pType, pTable, pDatabase )
	this.Destination.Type	:= UPPER(pType)
	this.Destination.Table	:= pTable

	if	argCount() >= 3
		this.Destination.Database	:= pDatabase
	else
		this.Destination.Database	:= null
	endif

	return


FUNCTION AddSchemaOptions(pObj, pKey, pValue)		// pObj		= obj.Source or obj.Destination  
																	// pKey		= ColNameHeader, Format, CharacterSet, DateTimeFormat, ...
																	// pValue	= (ColNameHeader)=true/false, (Format)=TabDelimited/CSVDelimited/FixedLength/Delimited(custom character) by ex. Delimited(;)
																	//				  (CharacterSet)=ANSI/OEM, ...
	pObj.Schema.Options.add( {pKey, pValue} )
	return



FUNCTION AddSchemaColumns(pObj, pColName, pFieldName, pFieldProperties)			// pObj					= obj.Source or obj.Destination
																										// pColName	= MUST BE : Col1 / Col2 / ... / Col10 / ... /Coln
																										// pFieldName			= attributed FieldName
																										// pFieldProperties	= CHAR, TEXT, FLOAT, INTEGER, Date, DateTime ....
																										//								optional can be specified the length like: CHAR width 25

	pObj.Schema.Columns.add( {pColName, pFieldName, pFieldProperties} )
	return

FUNCTION	CreateSchemaIni( pObj )				// parameter can be: this.source or this.destination
	local	x, i, oIni, pth, fName

	if	empty(pObj.Type) or NOT isCSV( pObj.Type )		// Create Schema.ini only for type: TXT, CSV, TEXT, DELIMITED 
		return
	endif
	
	pth		= pObj.Database
	pth	 	+= iif(right(pth,1)=='\', '', '\')
	fName		= pObj.Table
	
	oIni	= new ini(pth+'Schema.ini')						// Create or Open Schema.ini file
	
	if	oIni.GetSectionNames().scan(fname) > 0 AND pObj.Schema.Options.size == 0 AND pObj.Schema.Columns.size == 0  // EXIT IF no option or columns defined
		oIni.release()																																// AND section with the filename exist
		return
	endif

	if	pObj.Schema.Columns.size > 0 and oIni.GetSectionNames().scan(fname) > 0													// if fields are defined
		oIni.DeleteSection(fName)																												// will create a new fName section
	endif
	
	for i	= 1 to pObj.Schema.DefaultOptions.size																							// set default options
		oIni.SetValue( fName, pObj.Schema.DefaultOptions[i][1], ''+pObj.Schema.DefaultOptions[i][2] )
	endfor
	
	for i	= 1 to pObj.Schema.Options.size																									// default options can be overrided
		oIni.SetValue( fName, pObj.Schema.Options[i][1], ''+pObj.Schema.Options[i][2] )
	endfor
	
	for i	= 1 to pObj.Schema.Columns.size																									// set columns descriptions
		oIni.SetValue( fName, pObj.Schema.Columns[i][1], pObj.Schema.Columns[i][2]+' '+pObj.Schema.Columns[i][3] )
	endfor
	
	oIni.flush()																																	// save Schema.ini
	
	oIni.release()

	return



FUNCTION	ListSource( pNRows, pNCols )									// List nRows, nFields
	local	sRws, nRows, nCols
	
	this.CreateSchemaIni(this.Source)
	
	sRws	= this.SetRowset(this.Source)

	if	empty(sRws)
	
		private	n, pObj
		pObj	= this.Source
		n	= 'this.' + pObj.Name
		
		?;? '>> ListSource error', 'Object:   &n..Type =' at 20,  pObj.Type
		? '&n..dbs.active =' at 30, pObj.dbs.active
		? '&n..qry.active =' at 30, pObj.qry.active
		? '&n..qry.SQL    =' at 30, pObj.qry.SQL
		return
		
	endif
	
	nRows	= iif(argcount()<1, 0, pNRows)
	
	nCols	= iif(argcount()<2, sRws.fields.size, pNCols)
	nCols	= min(nCols, sRws.fields.size)
	
	?;? "Table:", sRws.tableName, '  fields structure'
	
	for i = 1 to nCols
		? str(i,5),
		??sRws.fields[i].FieldName
		??sRws.fields[i].Type		at 30
		??sRws.fields[i].Length		at 42
	endfor
	
	?;?'List ', nRows, 'rows   x', nCols, ' fields'
	?;?'Record'
	for j = 1 to nCols
		??sRws.fields[j].FieldName at(12*j),
	endfor

	i	= 1
	sRws.first()
	
	do
		? str(i,5),
		for j = 1 to nCols
			??	left(ltrim(''+sRws.fields[j].value),11) at 12*j,
		endfor
		
		if	nRows>0 and i++ >= nRows
			exit
		endif
		
	until not sRws.next()
	
	this.InactivateDataAccessObject()

	return



FUNCTION	Copy												// Copy Source to a new Destination
	local	sRws, dRws, rVal								// Source, Destination rowset

	rVal	= false
	
	this.CreateSchemaIni(this.Source)						// Create SCHEMA.INI if	necessary
	
	sRws	= this.SetRowset(this.Source)						// get rowset object for source

	if NOT empty(sRws)

		this.CreateSchemaIni( this.Destination )			// Create SCHEMA.INI if	necessary

		if	this.CreateTableFromSource( sRws )				// Create the new table with the structure of the source table

			dRws	= this.SetRowset( this.Destination )	// get rowset object for destination

			if	NOT empty(dRws)

				rVal	= this.CopyData(sRws, dRws)			// copy Source table into Destination table
				
			endif
			
		endif
		
	endif

	this.InactivateDataAccessObject()

	return	rVal
	
	
FUNCTION	Append
	local	sRws, dRws, rVal								// Source, Destination rowset

	rVal	= false
	this.CreateSchemaIni(this.Source)						// Create SCHEMA.INI if	necessary
	
	sRws	= this.SetRowset(this.Source)						// get rowset object for source
	
	if NOT empty(sRws)
	
		this.CreateSchemaIni( this.Destination )			// Create SCHEMA.INI if	necessary
		
		dRws	= this.SetRowset( this.Destination )	// get rowset object for destination

		if	NOT empty(dRws)
			rVal	= this.CopyData(sRws, dRws)			// copy the Source table into the Destination table
		endif
		
	endif
	
	this.InactivateDataAccessObject()
		
	return	rVal

FUNCTION	CreateTableFromSource(rs)							// rs	= Source table rowset
	private	x, dbs
	local		i, t, x, rVal
	
	rVal	= false
	
	x	= this.Destination.Table								
	if	type('x') == 'O'
		return	true
	endif
	
	this.SetRowset( this.Destination, true )
	
	dbs	= this.Destination.dbs
	
	if	dbs.active
	
		if	this.Destination.qry.active
			this.Destination.qry.active := false
		endif
		
		if	dbs.TableExists(this.Destination.Table)
				dbs.DropTable(this.Destination.Table)
		endif

		if	not isCSV(this.Destination.Type)
				x	= " ("												// create list of the fields definition
				for	i = 1 to rs.fields.size
					x	+= " " + rs.fields[i].FieldName + " "
					t	= upper(rs.fields[i].Type)
					t	= iif(t = "VARW",		"VARCHAR", t)
					t	= iif(t = "DOUBLE", 	"FLOAT",	  t)
					t	= iif(t = "BOOLEAN",	iif(isXLS(this.Destination.Type), "BIT", t), t)		// Excel datamaping for BOOLEAN is BIT
					x	+= t
					x	+= iif({"CHAR", "VARC", "VARW", "NUME"}.scan(left(t,4))>0, ;
							"(" + min(254, rs.fields[i].Length) + iif({"NUME"}.scan(left(t,4))>0, ","+rs.fields[i].decimalLength, "") ;
							+ ")", "")
					x	+= ","
				endfor
				x	= left(x, len(x)-1) + " )"
			
			
				try
					dbs.ExecuteSQL("CREATE TABLE " + this.Destination.Table + x)
					rVal	= true

				catch (exception e)
				endtry
				
				if	not dbs.TableExists(this.Destination.Table)
					try
						dbs.ExecuteSQL("CREATE TABLE '" + this.Destination.Table + "'" + x)
						rVal	= true
					catch (exception e)
					endtry
				endif
				
		else
		
				pth		= this.Destination.Database					// for type=CSV, TXT	create an empty file
				pth	 	+= iif(right(pth,1)=='\', '', '\')
				fName		= this.Destination.Table

				x	= new file()
				x.delete(pth+fName)
				rVal	= x.Create(pth+fName)
				x.close()
		endif

	endif
	
	return	rVal

FUNCTION	CopyData(rs, rd)
	local	i, fCount
	
	fCount	= min(rs.fields.size, rd.fields.size)
	
	if	rs.first()
	
		do
			rd.beginAppend()
			for	i = 1 to fCount
				rd.fields[i].value	:= rs.fields[i].value
			endfor
			rd.save()
		until not rs.next()
		
	endif
	
	return
	
FUNCTION	SetRowset
	parameter	pObj, dontTrigerError
	private	x, dbs, cDsn, cTbl, retVal
	
	x	= pObj.Table					// Can be Rowset object, SELECT ... , TableName, SheetName, FileName
	
	if	empty(x)
		return	null
	endif
	
	if	type('x') = 'O'
		if	{"ADOROWSET", "ROWSET"}.scan(x.baseClassName) > 0
			return x
		endif
 
	endif
	
	retVal	= null
	
	x	= pObj.Type
	
	if	isAdoType(x)
	
		dbs	= new AdoDatabase()
		
		cDsn	= pObj.Database

		do	case

		case	isCSV(x)							// cDsn = Folder (directory) where the file reside
				dbs.ConnectionString	:= 'Driver={Microsoft Text Driver (*.txt; *.csv)};Dbq=&cDsn.  ; Extensions=asc,csv,tab,txt;'

		case	isXLS(x)							// cDsn = the file.xlsx, .xls
				if	UPPER(substr(cDsn, rat('.',cDsn))) == '.XLSX'
					dbs.ConnectionString	:= 'Provider=Microsoft.ACE.OLEDB.16.0 ;   Data Source=&cDsn.  ; Extended Properties="Excel 12.0 Xml; HDR=YES" ;'
				else
					dbs.ConnectionString	:= 'Provider=Microsoft.ACE.OLEDB.16.0 ;   Data Source=&cDsn.  ; Extended Properties="Excel 8.0;HDR=YES" ;'
				endif
			
		otherwise
			if	type('cDsn') == 'C'
				dbs.ConnectionString	:= cDsn	// in pObj.Database is ConnectionString 
			else
				dbs	= cDsn						// in pObj.Database pointer to AdoDatabase object
			endif

		endcase
		
		try
			if	not dbs.active
				dbs.active	:= true
			endif
		catch (exception e)
		endtry
		
		qry	= new AdoQuery()
		
		if	dbs.active
		
			qry.database	= dbs
			qry.sql		= iif( isSqlSelect(pObj.Table),  pObj.Table,  "SELECT * FROM " + pObj.Table + "" )
			try
				qry.active	= true
			catch (exception e)
			endtry
			
		endif

	else					// using Local SQL
		
		dbs	= new DataBase()
		
		cDsn	= pObj.Database
		
		do	case
		
		case	empty(cDsn)
				dbs	:= _app.databases[1]
				
		case	type('cDsn') = 'C'
				dbs.DataBaseName	:= cDsn
				
		otherwise
				dbs	:= cDsn
		endcase
		
		try
			if	not dbs.active
				dbs.active	:= true
			endif
		catch (exception e)
		endtry

		qry	= new Query()
		
		if	dbs.active

			qry.database	:= dbs
			qry.sql			:= 	iif( isSqlSelect(pObj.Table),  pObj.Table,  "SELECT * FROM " + pObj.Table )
			
			try
				qry.active	= true
			catch (exception e)
			endtry
			
			if	not qry.active
				qry.sql			:= iif( isSqlSelect(pObj.Table),  pObj.Table,  "SELECT * FROM '" + pObj.Table + "'" )
				try
					qry.active	= true
				catch (exception e)
				endtry
			endif
			
		endif


	endif
	
	pObj.dbs = dbs
	pObj.qry	= qry
	
	retVal	= iif(qry.active, qry.rowset, null)
	
	if	empty(retVal) and dontTrigerError == false
	
		private	n
		n	= 'this.' + pObj.Name
		
		?;? '>> SetRowset error', 'Object:   &n..Type =' at 20,  pObj.Type
		? '&n..dbs.active =' at 30, pObj.dbs.active
		? '&n..qry.active =' at 30, pObj.qry.active
		? '&n..qry.SQL    =' at 30, pObj.qry.SQL
		
	endif
	
	return	retVal
	
	
	
FUNCTION	InactivateDataAccessObject()

	this.Inactivate(this.source)
	this.Inactivate(this.Destination)
	
	return

FUNCTION	Inactivate( pObj )
	private	x
	
	x	= pObj.Table
	if	type('x') <> 'O'
		try
			pObj.qry.active	:= false
			pObj.qry	:= null
		catch (exception e)
		endtry
	endif
	
	x	= pObj.Database
	if	type('x') <> 'O'
		try
			pObj.dbs.active	:= false
			pObj.dbs	:= null
		catch (exception e)
		endtry
	endif
	
	return

FUNCTION	Release

	this.InactivateDataAccessObject()
	release object this
	return	null
	return
endclass